@@ -20,11 +20,15 @@ require 'pg'
 describe PG::Connection do
 
 	before( :all ) do
-		@conn = setup_testing_db( "PG_Connection" )
+		@conn = setup_testing_db( described_class.name )
 	end
 
 	before( :each ) do
 		@conn.exec( 'BEGIN' ) unless example.metadata[:without_transaction]
+		if PG.respond_to?( :library_version )
+			@conn.exec_params %Q{SET application_name TO '%s'} %
+				[@conn.escape_string(example.description[0,60])]
+		end
 	end
 
 	after( :each ) do
@@ -41,7 +45,7 @@ describe PG::Connection do
 	#
 
 	it "can create a connection option string from a Hash of options" do
-		optstring = described_class.parse_connect_args( 
+		optstring = described_class.parse_connect_args(
 			:host => 'pgsql.example.com',
 			:dbname => 'db01',
 			'sslmode' => 'require'
@@ -54,14 +58,14 @@ describe PG::Connection do
 	end
 
 	it "can create a connection option string from positional parameters" do
-		optstring = described_class.parse_connect_args( 'pgsql.example.com', nil, '-c geqo=off', nil, 
+		optstring = described_class.parse_connect_args( 'pgsql.example.com', nil, '-c geqo=off', nil,
 		                                       'sales' )
 
 		optstring.should be_a( String )
 		optstring.should =~ /(^|\s)host='pgsql.example.com'/
 		optstring.should =~ /(^|\s)dbname='sales'/
 		optstring.should =~ /(^|\s)options='-c geqo=off'/
-		
+
 		optstring.should_not =~ /port=/
 		optstring.should_not =~ /tty=/
 	end
@@ -77,7 +81,8 @@ describe PG::Connection do
 	end
 
 	it "escapes single quotes and backslashes in connection parameters" do
-		described_class.parse_connect_args( "DB 'browser' \\" ).should == "host='DB \\'browser\\' \\\\'"
+		described_class.parse_connect_args( "DB 'browser' \\" ).
+			should =~ /host='DB \\'browser\\' \\\\'/
 
 	end
 
@@ -106,7 +111,7 @@ describe PG::Connection do
 		tmpconn.finish
 	end
 
-	it "connects using a hash of optional connection parameters" do
+	it "connects using a hash of optional connection parameters", :postgresql_90 do
 		tmpconn = described_class.connect(
 			:host => 'localhost',
 			:port => @port,
@@ -122,10 +127,10 @@ describe PG::Connection do
 		}.to raise_error( ArgumentError, /extra positional parameter/i )
 	end
 
-	it "can connect asynchronously" do
+	it "can connect asynchronously", :socket_io do
 		tmpconn = described_class.connect_start( @conninfo )
 		tmpconn.should be_a( described_class )
-		socket = IO.for_fd( tmpconn.socket )
+		socket = tmpconn.socket_io
 		status = tmpconn.connect_poll
 
 		while status != PG::PGRES_POLLING_OK
@@ -144,13 +149,13 @@ describe PG::Connection do
 		tmpconn.finish
 	end
 
-	it "can connect asynchronously for the duration of a block" do
+	it "can connect asynchronously for the duration of a block", :socket_io do
 		conn = nil
 
 		described_class.connect_start(@conninfo) do |tmpconn|
 			tmpconn.should be_a( described_class )
 			conn = tmpconn
-			socket = IO.for_fd(tmpconn.socket)
+			socket = tmpconn.socket_io
 			status = tmpconn.connect_poll
 
 			while status != PG::PGRES_POLLING_OK
@@ -172,6 +177,11 @@ describe PG::Connection do
 		conn.should be_finished()
 	end
 
+	it "raises proper error when sending fails" do
+		conn = described_class.connect_start( '127.0.0.1', 54320, "", "", "me", "xxxx", "somedb" )
+		expect{ conn.exec 'SELECT 1' }.to raise_error(PG::UnableToSend, /no connection/)
+	end
+
 	it "doesn't leave stale server connections after finish" do
 		described_class.connect(@conninfo).finish
 		sleep 0.5
@@ -211,8 +221,7 @@ describe PG::Connection do
 		From backend> T
 		}.gsub( /^\t{2}/, '' ).lstrip
 
-	unless RUBY_PLATFORM =~ /mswin|mingw/
-		it "trace and untrace client-server communication" do
+	it "trace and untrace client-server communication", :unix do
 			# be careful to explicitly close files so that the
 			# directory can be removed and we don't have to wait for
 			# the GC to run.
@@ -241,7 +250,6 @@ describe PG::Connection do
 
 			trace_data.should == expected_trace_output
 		end
-	end
 
 	it "allows a query to be cancelled" do
 		error = false
@@ -254,7 +262,42 @@ describe PG::Connection do
 		error.should == true
 	end
 
-	it "automatically rolls back a transaction started with described_class#transaction if an exception " +
+	it "can stop a thread that runs a blocking query with async_exec" do
+		start = Time.now
+		t = Thread.new do
+			@conn.async_exec( 'select pg_sleep(10)' )
+		end
+		sleep 0.1
+
+		t.kill
+		t.join
+		(Time.now - start).should < 10
+	end
+
+	it "should work together with signal handlers" do
+		signal_received = false
+		trap 'USR1' do
+			signal_received = true
+		end
+
+		Thread.new do
+			sleep 0.1
+			Process.kill("USR1", Process.pid)
+		end
+		@conn.exec("select pg_sleep(0.3)")
+		signal_received.should be_true
+
+		signal_received = false
+		Thread.new do
+			sleep 0.1
+			Process.kill("USR1", Process.pid)
+		end
+		@conn.async_exec("select pg_sleep(0.3)")
+		signal_received.should be_true
+	end
+
+
+	it "automatically rolls back a transaction started with Connection#transaction if an exception " +
 	   "is raised" do
 		# abort the per-example transaction so we can test our own
 		@conn.exec( 'ROLLBACK' )
@@ -273,6 +316,16 @@ describe PG::Connection do
 		res.ntuples.should == 0
 	end
 
+	it "returns the block result from Connection#transaction" do
+		# abort the per-example transaction so we can test our own
+		@conn.exec( 'ROLLBACK' )
+
+		res = @conn.transaction do
+			"transaction result"
+		end
+		res.should == "transaction result"
+	end
+
 	it "not read past the end of a large object" do
 		@conn.transaction do
 			oid = @conn.lo_create( 0 )
@@ -285,39 +338,58 @@ describe PG::Connection do
 	end
 
 
+	it "supports parameters passed to #exec (backward compatibility)" do
+		@conn.exec( "CREATE TABLE students ( name TEXT, age INTEGER )" )
+		@conn.exec( "INSERT INTO students VALUES( $1, $2 )", ['Wally', 8] )
+		@conn.exec( "INSERT INTO students VALUES( $1, $2 )", ['Sally', 6] )
+		@conn.exec( "INSERT INTO students VALUES( $1, $2 )", ['Dorothy', 4] )
+
+		res = @conn.exec( "SELECT name FROM students WHERE age >= $1", [6] )
+		res.values.should == [ ['Wally'], ['Sally'] ]
+	end
+
+	it "supports explicitly calling #exec_params" do
+		@conn.exec( "CREATE TABLE students ( name TEXT, age INTEGER )" )
+		@conn.exec( "INSERT INTO students VALUES( $1, $2 )", ['Wally', 8] )
+		@conn.exec( "INSERT INTO students VALUES( $1, $2 )", ['Sally', 6] )
+		@conn.exec( "INSERT INTO students VALUES( $1, $2 )", ['Dorothy', 4] )
+
+		res = @conn.exec_params( "SELECT name FROM students WHERE age >= $1", [6] )
+		res.values.should == [ ['Wally'], ['Sally'] ]
+	end
+
+
 	it "can wait for NOTIFY events" do
 		@conn.exec( 'ROLLBACK' )
 		@conn.exec( 'LISTEN woo' )
 
-		pid = fork do
+		t = Thread.new do
 			begin
 				conn = described_class.connect( @conninfo )
 				sleep 1
-				conn.exec( 'NOTIFY woo' )
+				conn.async_exec( 'NOTIFY woo' )
 			ensure
 				conn.finish
-				exit!
 			end
 		end
 
 		@conn.wait_for_notify( 10 ).should == 'woo'
 		@conn.exec( 'UNLISTEN woo' )
 
-		Process.wait( pid )
+		t.join
 	end
 
 	it "calls a block for NOTIFY events if one is given" do
 		@conn.exec( 'ROLLBACK' )
 		@conn.exec( 'LISTEN woo' )
 
-		pid = fork do
+		t = Thread.new do
 			begin
 				conn = described_class.connect( @conninfo )
 				sleep 1
-				conn.exec( 'NOTIFY woo' )
+				conn.async_exec( 'NOTIFY woo' )
 			ensure
 				conn.finish
-				exit!
 			end
 		end
 
@@ -328,7 +400,7 @@ describe PG::Connection do
 
 		@conn.exec( 'UNLISTEN woo' )
 
-		Process.wait( pid )
+		t.join
 	end
 
 	it "doesn't collapse sequential notifications" do
@@ -337,20 +409,15 @@ describe PG::Connection do
 		@conn.exec( 'LISTEN war' )
 		@conn.exec( 'LISTEN woz' )
 
-		pid = fork do
-			begin
-				conn = described_class.connect( @conninfo )
-				conn.exec( 'NOTIFY woo' )
-				conn.exec( 'NOTIFY war' )
-				conn.exec( 'NOTIFY woz' )
-			ensure
-				conn.finish
-				exit!
-			end
+		begin
+			conn = described_class.connect( @conninfo )
+			conn.exec( 'NOTIFY woo' )
+			conn.exec( 'NOTIFY war' )
+			conn.exec( 'NOTIFY woz' )
+		ensure
+			conn.finish
 		end
 
-		Process.wait( pid )
-
 		channels = []
 		3.times do
 			channels << @conn.wait_for_notify( 2 )
@@ -369,19 +436,13 @@ describe PG::Connection do
 		@conn.exec( 'ROLLBACK' )
 		@conn.exec( 'LISTEN woo' )
 
-		pid = fork do
-			begin
-				conn = described_class.connect( @conninfo )
-				conn.exec( 'NOTIFY woo' )
-			ensure
-				conn.finish
-				exit!
-			end
+		begin
+			conn = described_class.connect( @conninfo )
+			conn.exec( 'NOTIFY woo' )
+		ensure
+			conn.finish
 		end
 
-		# Wait for the forked child to send the notification
-		Process.wait( pid )
-
 		# Cause the notification to buffer, but not be read yet
 		@conn.exec( 'SELECT 1' )
 
@@ -389,6 +450,21 @@ describe PG::Connection do
 		@conn.exec( 'UNLISTEN woo' )
 	end
 
+	it "can receive notices while waiting for NOTIFY without exceeding the timeout", :postgresql_90 do
+		notices = []
+		@conn.set_notice_processor do |msg|
+			notices << [msg, Time.now]
+		end
+		st = Time.now
+		@conn.send_query "SELECT pg_sleep(0.5); do $$ BEGIN RAISE NOTICE 'woohoo'; END; $$ LANGUAGE plpgsql;"
+		@conn.wait_for_notify( 1 ).should be_nil
+		notices.first.should_not be_nil
+		et = Time.now
+		(et - notices.first[1]).should >= 0.4
+		(et - st).should >= 0.9
+		(et - st).should < 1.4
+	end
+
 	it "yields the result if block is given to exec" do
 		rval = @conn.exec( "select 1234::int as a union select 5678::int as a" ) do |result|
 			values = []
@@ -404,6 +480,101 @@ describe PG::Connection do
 		rval.should include( '5678', '1234' )
 	end
 
+	it "can process #copy_data output queries" do
+		rows = []
+		res2 = @conn.copy_data( "COPY (SELECT 1 UNION ALL SELECT 2) TO STDOUT" ) do |res|
+			res.result_status.should == PG::PGRES_COPY_OUT
+			res.nfields.should == 1
+			while row=@conn.get_copy_data
+				rows << row
+			end
+		end
+		rows.should == ["1\n", "2\n"]
+		res2.result_status.should == PG::PGRES_COMMAND_OK
+		verify_clean_exec_status
+	end
+
+	it "can handle incomplete #copy_data output queries" do
+		expect {
+			@conn.copy_data( "COPY (SELECT 1 UNION ALL SELECT 2) TO STDOUT" ) do |res|
+				@conn.get_copy_data
+			end
+		}.to raise_error(PG::NotAllCopyDataRetrieved, /Not all/)
+		verify_clean_exec_status
+	end
+
+	it "can handle client errors in #copy_data for output" do
+		expect {
+			@conn.copy_data( "COPY (SELECT 1 UNION ALL SELECT 2) TO STDOUT" ) do
+				raise "boom"
+			end
+		}.to raise_error(RuntimeError, "boom")
+		verify_clean_exec_status
+	end
+
+	it "can handle server errors in #copy_data for output" do
+		@conn.exec "ROLLBACK"
+		@conn.transaction do
+			@conn.exec( "CREATE FUNCTION errfunc() RETURNS int AS $$ BEGIN RAISE 'test-error'; END; $$ LANGUAGE plpgsql;" )
+			expect {
+				@conn.copy_data( "COPY (SELECT errfunc()) TO STDOUT" ) do |res|
+					while @conn.get_copy_data
+					end
+				end
+			}.to raise_error(PG::Error, /test-error/)
+		end
+		verify_clean_exec_status
+	end
+
+	it "can process #copy_data input queries" do
+		@conn.exec( "CREATE TEMP TABLE copytable (col1 TEXT)" )
+		res2 = @conn.copy_data( "COPY copytable FROM STDOUT" ) do |res|
+			res.result_status.should == PG::PGRES_COPY_IN
+			res.nfields.should == 1
+			@conn.put_copy_data "1\n"
+			@conn.put_copy_data "2\n"
+		end
+		res2.result_status.should == PG::PGRES_COMMAND_OK
+
+		verify_clean_exec_status
+
+		res = @conn.exec( "SELECT * FROM copytable ORDER BY col1" )
+		res.values.should == [["1"], ["2"]]
+	end
+
+	it "can handle client errors in #copy_data for input" do
+		@conn.exec "ROLLBACK"
+		@conn.transaction do
+			@conn.exec( "CREATE TEMP TABLE copytable (col1 TEXT)" )
+			expect {
+				@conn.copy_data( "COPY copytable FROM STDOUT" ) do |res|
+					raise "boom"
+				end
+			}.to raise_error(RuntimeError, "boom")
+		end
+		verify_clean_exec_status
+	end
+
+	it "can handle server errors in #copy_data for input" do
+		@conn.exec "ROLLBACK"
+		@conn.transaction do
+			@conn.exec( "CREATE TEMP TABLE copytable (col1 INT)" )
+			expect {
+				@conn.copy_data( "COPY copytable FROM STDOUT" ) do |res|
+					@conn.put_copy_data "xyz\n"
+				end
+			}.to raise_error(PG::Error, /invalid input syntax for integer/)
+		end
+		verify_clean_exec_status
+	end
+
+	it "should raise an error for non copy statements in #copy_data" do
+		expect {
+			@conn.copy_data( "SELECT 1" ){}
+		}.to raise_error(ArgumentError, /no COPY/)
+
+		verify_clean_exec_status
+	end
 
 	it "correctly finishes COPY queries passed to #async_exec" do
 		@conn.async_exec( "COPY (SELECT 1 UNION ALL SELECT 2) TO STDOUT" )
@@ -424,16 +595,17 @@ describe PG::Connection do
 
 
 	it "described_class#block shouldn't block a second thread" do
+		start = Time.now
 		t = Thread.new do
 			@conn.send_query( "select pg_sleep(3)" )
 			@conn.block
 		end
 
-		# :FIXME: There's a race here, but hopefully it's pretty small.
+		sleep 0.5
 		t.should be_alive()
-
 		@conn.cancel
 		t.join
+		(Time.now - start).should < 3
 	end
 
 	it "described_class#block should allow a timeout" do
@@ -505,14 +677,14 @@ describe PG::Connection do
 	end
 
 
-	it "can connect asynchronously" do
+	it "can connect asynchronously", :socket_io do
 		serv = TCPServer.new( '127.0.0.1', 54320 )
 		conn = described_class.connect_start( '127.0.0.1', 54320, "", "", "me", "xxxx", "somedb" )
-		conn.connect_poll.should == PG::PGRES_POLLING_WRITING
-		select( nil, [IO.for_fd(conn.socket)], nil, 0.2 )
+		[PG::PGRES_POLLING_WRITING, PG::CONNECTION_OK].should include conn.connect_poll
+		select( nil, [conn.socket_io], nil, 0.2 )
 		serv.close
 		if conn.connect_poll == PG::PGRES_POLLING_READING
-			select( [IO.for_fd(conn.socket)], nil, nil, 0.2 )
+			select( [conn.socket_io], nil, nil, 0.2 )
 		end
 		conn.connect_poll.should == PG::PGRES_POLLING_FAILED
 	end
@@ -531,9 +703,36 @@ describe PG::Connection do
 		conn = PG.connect( @conninfo )
 
 		conn.finish
-		expect { conn.finish }.to raise_error( PG::Error, /connection is closed/i )
+		expect { conn.finish }.to raise_error( PG::ConnectionBad, /connection is closed/i )
 	end
 
+	it "closes the IO fetched from #socket_io when the connection is closed", :without_transaction, :socket_io do
+		conn = PG.connect( @conninfo )
+		io = conn.socket_io
+		conn.finish
+		io.should be_closed()
+		expect { conn.socket_io }.to raise_error( PG::ConnectionBad, /connection is closed/i )
+	end
+
+	it "closes the IO fetched from #socket_io when the connection is reset", :without_transaction, :socket_io do
+		conn = PG.connect( @conninfo )
+		io = conn.socket_io
+		conn.reset
+		io.should be_closed()
+		conn.socket_io.should_not equal( io )
+		conn.finish
+	end
+
+	it "block should raise ConnectionBad for a closed connection" do
+		serv = TCPServer.new( '127.0.0.1', 54320 )
+		conn = described_class.connect_start( '127.0.0.1', 54320, "", "", "me", "xxxx", "somedb" )
+		while [PG::CONNECTION_STARTED, PG::CONNECTION_MADE].include?(conn.connect_poll)
+			sleep 0.1
+		end
+		serv.close
+		expect{ conn.block }.to raise_error(PG::ConnectionBad, /server closed the connection unexpectedly/)
+		expect{ conn.block }.to raise_error(PG::ConnectionBad, /can't get socket descriptor/)
+	end
 
 	context "under PostgreSQL 9", :postgresql_90 do
 
@@ -541,20 +740,31 @@ describe PG::Connection do
 			pending "only works with a PostgreSQL >= 9.0 server" if @conn.server_version < 9_00_00
 		end
 
+		it "sets the fallback_application_name on new connections" do
+			conn_string = PG::Connection.parse_connect_args( 'dbname=test' )
+			connection_string_should_contain_application_name(conn_string, $0)
+		end
+
+		it "sets a shortened fallback_application_name on new connections" do
+			old_0 = $0
+			begin
+				$0 = "/this/is/a/very/long/path/with/many/directories/to/our/beloved/ruby"
+				conn_string = PG::Connection.parse_connect_args( 'dbname=test' )
+				connection_string_should_contain_application_name(conn_string, $0)
+			ensure
+				$0 = old_0
+			end
+		end
+
 		it "calls the block supplied to wait_for_notify with the notify payload if it accepts " +
 		    "any number of arguments" do
 
 			@conn.exec( 'ROLLBACK' )
 			@conn.exec( 'LISTEN knees' )
 
-			pid = fork do
-				conn = described_class.connect( @conninfo )
-				conn.exec( %Q{NOTIFY knees, 'skirt and boots'} )
-				conn.finish
-				exit!
-			end
-
-			Process.wait( pid )
+			conn = described_class.connect( @conninfo )
+			conn.exec( %Q{NOTIFY knees, 'skirt and boots'} )
+			conn.finish
 
 			event, pid, msg = nil
 			@conn.wait_for_notify( 10 ) do |*args|
@@ -571,14 +781,9 @@ describe PG::Connection do
 			@conn.exec( 'ROLLBACK' )
 			@conn.exec( 'LISTEN knees' )
 
-			pid = fork do
-				conn = described_class.connect( @conninfo )
-				conn.exec( %Q{NOTIFY knees} )
-				conn.finish
-				exit!
-			end
-
-			Process.wait( pid )
+			conn = described_class.connect( @conninfo )
+			conn.exec( %Q{NOTIFY knees} )
+			conn.finish
 
 			event, pid = nil
 			@conn.wait_for_notify( nil ) do |*args|
@@ -594,14 +799,9 @@ describe PG::Connection do
 			@conn.exec( 'ROLLBACK' )
 			@conn.exec( 'LISTEN knees' )
 
-			pid = fork do
-				conn = described_class.connect( @conninfo )
-				conn.exec( %Q{NOTIFY knees} )
-				conn.finish
-				exit!
-			end
-
-			Process.wait( pid )
+			conn = described_class.connect( @conninfo )
+			conn.exec( %Q{NOTIFY knees} )
+			conn.finish
 
 			payload = :notnil
 			@conn.wait_for_notify( nil ) do |*args|
@@ -618,14 +818,9 @@ describe PG::Connection do
 			@conn.exec( 'ROLLBACK' )
 			@conn.exec( 'LISTEN knees' )
 
-			pid = fork do
-				conn = described_class.connect( @conninfo )
-				conn.exec( %Q{NOTIFY knees, 'skirt and boots'} )
-				conn.finish
-				exit!
-			end
-
-			Process.wait( pid )
+			conn = described_class.connect( @conninfo )
+			conn.exec( %Q{NOTIFY knees, 'skirt and boots'} )
+			conn.finish
 
 			event, pid, msg = nil
 			@conn.wait_for_notify( 10 ) do |arg1, arg2|
@@ -644,14 +839,9 @@ describe PG::Connection do
 			@conn.exec( 'ROLLBACK' )
 			@conn.exec( 'LISTEN knees' )
 
-			pid = fork do
-				conn = described_class.connect( @conninfo )
-				conn.exec( %Q{NOTIFY knees, 'skirt and boots'} )
-				conn.finish
-				exit!
-			end
-
-			Process.wait( pid )
+			conn = described_class.connect( @conninfo )
+			conn.exec( %Q{NOTIFY knees, 'skirt and boots'} )
+			conn.finish
 
 			notification_received = false
 			@conn.wait_for_notify( 10 ) do
@@ -668,14 +858,9 @@ describe PG::Connection do
 			@conn.exec( 'ROLLBACK' )
 			@conn.exec( 'LISTEN knees' )
 
-			pid = fork do
-				conn = described_class.connect( @conninfo )
-				conn.exec( %Q{NOTIFY knees, 'skirt and boots'} )
-				conn.finish
-				exit!
-			end
-
-			Process.wait( pid )
+			conn = described_class.connect( @conninfo )
+			conn.exec( %Q{NOTIFY knees, 'skirt and boots'} )
+			conn.finish
 
 			event, pid, msg = nil
 			@conn.wait_for_notify( 10 ) do |arg1, arg2, arg3|
@@ -726,22 +911,71 @@ describe PG::Connection do
 
 	end
 
-	context "multinationalization support", :ruby_19 do
+	context "under PostgreSQL 9.2 client library", :postgresql_92 do
+		describe "set_single_row_mode" do
 
-		it "should return the same bytes in text format that are sent as inline text" do
-			binary_file   = File.join(Dir.pwd, 'spec/data', 'random_binary_data')
-			in_bytes      = File.open(binary_file, 'r:ASCII-8BIT').read
-			escaped_bytes = described_class.escape_bytea( in_bytes )
-			out_bytes     = nil
-
-			@conn.transaction do |conn|
-				conn.exec("SET standard_conforming_strings=on")
-				res = conn.exec("VALUES ('#{escaped_bytes}'::bytea)", [], 0)
-				out_bytes = described_class.unescape_bytea( res[0]['column1'] )
+			it "raises an error when called at the wrong time" do
+				expect {
+					@conn.set_single_row_mode
+				}.to raise_error(PG::Error)
 			end
 
-			out_bytes.should == in_bytes
+			it "should work in single row mode" do
+				@conn.send_query( "SELECT generate_series(1,10)" )
+				@conn.set_single_row_mode
+
+				results = []
+				loop do
+					@conn.block
+					res = @conn.get_result or break
+					results << res
+				end
+				results.length.should == 11
+				results[0..-2].each do |res|
+					res.result_status.should == PG::PGRES_SINGLE_TUPLE
+					values = res.field_values('generate_series')
+					values.length.should == 1
+					values.first.to_i.should > 0
+				end
+				results.last.result_status.should == PG::PGRES_TUPLES_OK
+				results.last.ntuples.should == 0
+			end
+
+			it "should receive rows before entire query is finished" do
+				@conn.send_query( "SELECT generate_series(0,999), NULL UNION ALL SELECT 1000, pg_sleep(1);" )
+				@conn.set_single_row_mode
+
+				start_time = Time.now
+				first_row_time = nil
+				loop do
+					res = @conn.get_result or break
+					res.check
+					first_row_time = Time.now unless first_row_time
+				end
+				(Time.now - start_time).should >= 1.0
+				(first_row_time - start_time).should < 1.0
+			end
+
+			it "should receive rows before entire query fails" do
+				@conn.exec( "CREATE FUNCTION errfunc() RETURNS int AS $$ BEGIN RAISE 'test-error'; END; $$ LANGUAGE plpgsql;" )
+				@conn.send_query( "SELECT generate_series(0,999), NULL UNION ALL SELECT 1000, errfunc();" )
+				@conn.set_single_row_mode
+
+				first_result = nil
+				expect do
+					loop do
+						res = @conn.get_result or break
+						res.check
+						first_result ||= res
+					end
+				end.to raise_error(PG::Error)
+				first_result.kind_of?(PG::Result).should be_true
+				first_result.result_status.should == PG::PGRES_SINGLE_TUPLE
+			end
 		end
+	end
+
+	context "multinationalization support", :ruby_19 do
 
 		describe "rubyforge #22925: m17n support" do
 			it "should return results in the same encoding as the client (iso-8859-1)" do
@@ -812,17 +1046,53 @@ describe PG::Connection do
 			end
 
 			it "uses the client encoding for escaped string" do
-				original = "string to escape".force_encoding( "euc-jp" )
+				original = "string to\0 escape".force_encoding( "iso8859-1" )
 				@conn.set_client_encoding( "euc_jp" )
 				escaped  = @conn.escape( original )
 				escaped.encoding.should == Encoding::EUC_JP
+				escaped.should == "string to"
 			end
 
-			it "escapes string as literal" do
-				original = "string to\0 escape"
+			it "uses the client encoding for escaped literal", :postgresql_90 do
+				original = "string to\0 escape".force_encoding( "iso8859-1" )
+				@conn.set_client_encoding( "euc_jp" )
 				escaped  = @conn.escape_literal( original )
+				escaped.encoding.should == Encoding::EUC_JP
 				escaped.should == "'string to'"
 			end
+
+			it "uses the client encoding for escaped identifier", :postgresql_90 do
+				original = "string to\0 escape".force_encoding( "iso8859-1" )
+				@conn.set_client_encoding( "euc_jp" )
+				escaped  = @conn.escape_identifier( original )
+				escaped.encoding.should == Encoding::EUC_JP
+				escaped.should == "\"string to\""
+			end
+
+			it "uses the client encoding for quote_ident" do
+				original = "string to\0 escape".force_encoding( "iso8859-1" )
+				@conn.set_client_encoding( "euc_jp" )
+				escaped  = @conn.quote_ident( original )
+				escaped.encoding.should == Encoding::EUC_JP
+				escaped.should == "\"string to\""
+			end
+
+			it "uses the previous string encoding for escaped string" do
+				original = "string to\0 escape".force_encoding( "iso8859-1" )
+				@conn.set_client_encoding( "euc_jp" )
+				escaped  = described_class.escape( original )
+				escaped.encoding.should == Encoding::ISO8859_1
+				escaped.should == "string to"
+			end
+
+			it "uses the previous string encoding for quote_ident" do
+				original = "string to\0 escape".force_encoding( "iso8859-1" )
+				@conn.set_client_encoding( "euc_jp" )
+				escaped  = described_class.quote_ident( original )
+				escaped.encoding.should == Encoding::ISO8859_1
+				escaped.should == "\"string to\""
+			end
+
 		end
 
 
@@ -852,11 +1122,11 @@ describe PG::Connection do
 			it "allows users of the async interface to set the client_encoding to the default_internal" do
 				begin
 					prev_encoding = Encoding.default_internal
-					Encoding.default_internal = Encoding::KOI8_U
+					Encoding.default_internal = Encoding::KOI8_R
 
 					@conn.set_default_encoding
 
-					@conn.internal_encoding.should == Encoding::KOI8_U
+					@conn.internal_encoding.should == Encoding::KOI8_R
 				ensure
 					Encoding.default_internal = prev_encoding
 				end
@@ -886,5 +1156,91 @@ describe PG::Connection do
 			conn.finish if conn
 		end
 
+		it "receives properly encoded messages in the notice callbacks", :postgresql_90 do
+			[:receiver, :processor].each do |kind|
+				notices = []
+				@conn.internal_encoding = 'utf-8'
+				if kind == :processor
+					@conn.set_notice_processor do |msg|
+						notices << msg
+					end
+				else
+					@conn.set_notice_receiver do |result|
+						notices << result.error_message
+					end
+				end
+
+				3.times do
+					@conn.exec "do $$ BEGIN RAISE NOTICE '世界線航跡蔵'; END; $$ LANGUAGE plpgsql;"
+				end
+
+				notices.length.should == 3
+				notices.each do |notice|
+					notice.should =~ /^NOTICE:.*世界線航跡蔵/
+					notice.encoding.should == Encoding::UTF_8
+				end
+				@conn.set_notice_receiver
+				@conn.set_notice_processor
+			end
+		end
+
+		it "receives properly encoded text from wait_for_notify", :postgresql_90 do
+			@conn.internal_encoding = 'utf-8'
+			@conn.exec( 'ROLLBACK' )
+			@conn.exec( 'LISTEN "Möhre"' )
+			@conn.exec( %Q{NOTIFY "Möhre", '世界線航跡蔵'} )
+			event, pid, msg = nil
+			@conn.wait_for_notify( 10 ) do |*args|
+				event, pid, msg = *args
+			end
+			@conn.exec( 'UNLISTEN "Möhre"' )
+
+			event.should == "Möhre"
+			event.encoding.should == Encoding::UTF_8
+			msg.should == '世界線航跡蔵'
+			msg.encoding.should == Encoding::UTF_8
+		end
+
+		it "returns properly encoded text from notifies", :postgresql_90 do
+			@conn.internal_encoding = 'utf-8'
+			@conn.exec( 'ROLLBACK' )
+			@conn.exec( 'LISTEN "Möhre"' )
+			@conn.exec( %Q{NOTIFY "Möhre", '世界線航跡蔵'} )
+			@conn.exec( 'UNLISTEN "Möhre"' )
+
+			notification = @conn.notifies
+			notification[:relname].should == "Möhre"
+			notification[:relname].encoding.should == Encoding::UTF_8
+			notification[:extra].should == '世界線航跡蔵'
+			notification[:extra].encoding.should == Encoding::UTF_8
+			notification[:be_pid].should > 0
+		end
+	end
+
+	context "OS thread support", :ruby_19 do
+		it "Connection#exec shouldn't block a second thread" do
+			t = Thread.new do
+				@conn.exec( "select pg_sleep(1)" )
+			end
+
+			sleep 0.5
+			t.should be_alive()
+			t.join
+		end
+
+		it "Connection.new shouldn't block a second thread" do
+			serv = nil
+			t = Thread.new do
+				serv = TCPServer.new( '127.0.0.1', 54320 )
+				expect {
+					described_class.new( '127.0.0.1', 54320, "", "", "me", "xxxx", "somedb" )
+				}.to raise_error(PG::ConnectionBad, /server closed the connection unexpectedly/)
+			end
+
+			sleep 0.5
+			t.should be_alive()
+			serv.close
+			t.join
+		end
 	end
 end
